import re
from datetime import datetime, date

from pydantic import BaseModel
from sqlmodel import Field, Relationship
import pandas as pd

from gisaf.models.models_base import Model
from gisaf.models.survey import Surveyor, Equipment
from gisaf.models.project import Project
from gisaf.models.metadata import gisaf_admin, gisaf_survey


re_file_import_record_date_expr = '^(\S+)-(\d\d\d\d)-(\d\d)-(\d\d).*$'
re_file_import_record_date_combined_expr = '^\S+-(\d\d\d\d-\d\d-\d\d).*$'
re_file_import_record_date = re.compile(re_file_import_record_date_expr)


class BadSurveyFileName(Exception):
    pass


def get_file_import_date(record) -> date:
    """
    Utility function that returns the date of survey from the file name,
    if it matches the convention for CSV survey files.
    Return exception otherwise.
    """
    fname_search = re_file_import_record_date.search(record['name'])
    if not fname_search:
        raise BadSurveyFileName(
            'The file name is not OK '
            '(format should be: "PPP-DESCRIPTION-YYYY-MM-DD", '
            'PPP being the project name, DESCRITION is optional and discarded)'
        )
    return date(day=int(fname_search.group(4)),
                month=int(fname_search.group(3)),
                year=int(fname_search.group(2)))


class FileImport(Model, table=True):
    """
    Files to import or imported in the DB.
    Give either url or path.
    """
    __tablename__: str = 'file_import' # type: ignore
    __table_args__ = gisaf_admin.table_args

    id: int | None = Field(default=None, primary_key=True)
    url: str
    ## TODO: Deprecate FileImport.path, in favour of dir + name
    path: str
    dir: str
    name: str
    md5: str
    time: datetime
    comment: str
    status: str
    store: str
    basket: str
    project_id: int = Field(foreign_key=gisaf_admin.table('project.id'))
    project: Project = Relationship()
    # ALTER TABLE gisaf_admin.file_import add column  project_id INT REFERENCES gisaf_admin.project;
    surveyor_id: int = Field(foreign_key=gisaf_survey.table('surveyor.id'))
    surveyor: Surveyor = Relationship()
    # ALTER TABLE gisaf_admin.file_import add column surveyor_id INT REFERENCES gisaf_survey.surveyor;
    equipment_id: int = Field(foreign_key=gisaf_survey.table('equipment.id'))
    equipment: Equipment = Relationship()
    # ALTER TABLE gisaf_admin.file_import add column equipment_id INT REFERENCES gisaf_survey.equipment;

    def __str__(self):
        return f'{self.path:s} for project id {self.project_id}'

    def __repr__(self):
        return f'<gisaf.misc.FileImport (gisaf_admin.file_import) {self.path}>'

    @classmethod
    def selectinload(cls):
        return [cls.project, cls.surveyor, cls.equipment]

    ## XXX: was used in Flask
    # def set_import_time(self):
    #     self.time = datetime.now()
    #     db.session.commit()

    @classmethod
    async def get_df(cls, *args, **kwargs):
        """
        Add a column 'date' based on the file name
        """
        df = await super().get_df(*args, **kwargs)
        dates = df['name'].str.extract(re_file_import_record_date_combined_expr)
        df['date'] = pd.to_datetime(dates[0], format='%Y-%m-%d')
        return df


    #def get_parent_dir(self):
    #    split_path = self.path.split(os_path.sep)
    #    if len(split_path) == 1:
    #        return None
    #    else:
    #        return split_path[-2]

    #def get_absolute_path(self):
    #    return os_path.join(conf.shapefiles_import_path, self.path)

    #def exists_in_file_system(self):
    #    """
    #    Check if the file exists
    #    :return:
    #    """
    #    return os_path.exists(self.get_absolute_path())


class FeatureImportData(Model, table=True):
    """
    Keep track of imported data, typically from shapefiles
    """
    __tablename__: str = 'feature_import_data' # type: ignore
    __table_args__ = gisaf_admin.table_args

    id: int | None = Field(default=None, primary_key=True)
    store: str = Field(index=True)
    feature_id: int = Field(index=True)
    orig_id: int
    time: datetime
    origin: str
    file_path: str
    file_md5: str


class BasketFile(BaseModel):
    id: int
    dir: int
    name: int
    url: int
    md5: int
    time: datetime
    comment: int
    status: int
    store: int
    project: int
    surveyor: int
    equipment: int
    import_result: int


class BasketNameOnly(BaseModel):
    name: str


class Basket(BasketNameOnly):
    files: list[BasketFile]
    columns: list[str]
    uploadFields: list[str]
    projects: list[str]


class BasketImportResult(BaseModel):
    time: datetime
    message: str
    details: str

class AdminBasketFile(BaseModel):
    id: int
    dir: str
    name: str
    url: str
    md5: str
    time: datetime
    comment: str
    status: str
    store: str
    project: str
    surveyor: str
    equipment: str
    import_result: str


class AdminBasket(BaseModel):
    name: str
    files: list[FileImport]
    columns: list[str]
    uploadFields: list[str]
    projects: list[str] = []
